﻿#include "doors.h"
#include <QtXml/qdom.h>
#include <QFile>
#include <QDebug>
#include "../common/sources/system/cgraphics2d.h"
#include "../common/sources/system/filecache.h"
#include "player.h"
#include "mouse.h"
#include "tiles.h"
#include "game.h"
#include "timer.h"

CDoors   doors;

//----------------------------------------------------------------------------
CDoors::CDoors()
{

}

//----------------------------------------------------------------------------
// table pour la partie gauche de l'encadrement
SDoorElem   tabLeftFrame[WALL_TABLE_HEIGHT][WALL_TABLE_WIDTH] =
{
    // Rang 3 (le plus loin)
    {
        {"", 0, 0, 0, 0, 0, 0},  // 03
        {"Frame_Far_Left_0.png",  16, 59,  16, 59,  31, 101}, // 13
        {"Frame_Left_0.png",      82, 60,  82, 60,  91, 101}, // 23
        {"Frame_Far_Left_0.png", 141, 60, 147, 60, 150, 102}, // 33
        {"", 0, 0, 0, 0, 0, 0}   // 43
    },

    // Rang 2
    {
        {"", 0, 0, 0, 0, 0, 0},  // 02
        {"", 0, 0, 0, 0, 0, 0},  // 12
        {"Frame_Left_1.png", 65, 54, 65, 54, 82, 118}, // 22
        {"", 0, 0, 0, 0, 0, 0},  // 32
        {"", 0, 0, 0, 0, 0, 0}   // 42
    },

    // Rang 1
    {
        {"", 0, 0, 0, 0, 0, 0},  // 01
        {"", 0, 0, 0, 0, 0, 0},  // 11
        {"Frame_Left_2.png", 44, 46, 44, 46, 68, 139}, // 21
        {"", 0, 0, 0, 0, 0, 0},  // 31
        {"", 0, 0, 0, 0, 0, 0}   // 41
    },

    // Rang 0 (le plus proche)
    {
        {"", 0, 0, 0, 0, 0, 0},  // 00
        {"", 0, 0, 0, 0, 0, 0},  // 10
        {"", 0, 0, 0, 0, 0, 0},  // 20
        {"", 0, 0, 0, 0, 0, 0},  // 30
        {"", 0, 0, 0, 0, 0, 0}   // 40
    }
};

//----------------------------------------------------------------------------
// table pour la partie droite de l'encadrement
SDoorElem   tabRightFrame[WALL_TABLE_HEIGHT][WALL_TABLE_WIDTH] =
{
    // Rang 3 (le plus loin)
    {
        {"", 0, 0, 0, 0, 0, 0},  // 03
        {"Frame_Far_Left_0.png",  67, 60,  73, 60,  76, 102}, // 13
        {"Frame_Left_0.png",     132, 60, 132, 60, 141, 101}, // 23
        {"Frame_Far_Left_0.png", 192, 59, 192, 59, 207, 101}, // 33
        {"", 0, 0, 0, 0, 0, 0}   // 43
    },

    // Rang 2
    {
        {"", 0, 0, 0, 0, 0, 0},  // 02
        {"", 0, 0, 0, 0, 0, 0},  // 12
        {"Frame_Left_1.png", 141, 54, 141, 54, 158, 118}, // 22
        {"", 0, 0, 0, 0, 0, 0},  // 32
        {"", 0, 0, 0, 0, 0, 0}   // 42
    },

    // Rang 1
    {
        {"", 0, 0, 0, 0, 0, 0},  // 01
        {"", 0, 0, 0, 0, 0, 0},  // 11
        {"Frame_Left_2.png", 155, 46, 155, 46, 179, 139}, // 21
        {"", 0, 0, 0, 0, 0, 0},  // 31
        {"", 0, 0, 0, 0, 0, 0}   // 41
    },

    // Rang 0 (le plus proche)
    {
        {"", 0, 0, 0, 0, 0, 0},  // 00
        {"", 0, 0, 0, 0, 0, 0},  // 10
        {"", 0, 0, 0, 0, 0, 0},  // 20
        {"", 0, 0, 0, 0, 0, 0},  // 30
        {"", 0, 0, 0, 0, 0, 0}   // 40
    }
};

//----------------------------------------------------------------------------
// table pour la partie haute de l'encadrement
SDoorElem   tabTopFrame[WALL_TABLE_HEIGHT][WALL_TABLE_WIDTH] =
{
    // Rang 3 (le plus loin)
    {
        {"", 0, 0, 0, 0, 0, 0},  // 03
        {"", 0, 0, 0, 0, 0, 0},  // 13
        {"", 0, 0, 0, 0, 0, 0},  // 23
        {"", 0, 0, 0, 0, 0, 0},  // 33
        {"", 0, 0, 0, 0, 0, 0}   // 43
    },

    // Rang 2
    {
        {"", 0, 0, 0, 0, 0, 0},  // 02
        {"Frame_Top_1.png",   0, 55,   0, 55,  69, 57}, // 12
        {"Frame_Top_1.png",  77, 54,  77, 54, 146, 56}, // 22
        {"Frame_Top_1.png", 154, 55, 154, 55, 223, 57}, // 32
        {"", 0, 0, 0, 0, 0, 0}   // 42
    },

    // Rang 1
    {
        {"", 0, 0, 0, 0, 0, 0},  // 01
        {"Frame_Top_2.png", -53, 47,   0, 47,  31, 50}, // 11
        {"Frame_Top_2.png",  61, 45,  61, 45, 162, 48}, // 21
        {"Frame_Top_2.png", 176, 47, 192, 47, 223, 50}, // 31
        {"", 0, 0, 0, 0, 0, 0}   // 41
    },

    // Rang 0 (le plus proche)
    {
        {"", 0, 0, 0, 0, 0, 0},  // 00
        {"", 0, 0, 0, 0, 0, 0},  // 10
        {"", 0, 0, 0, 0, 0, 0},  // 20
        {"", 0, 0, 0, 0, 0, 0},  // 30
        {"", 0, 0, 0, 0, 0, 0}   // 40
    }
};

//----------------------------------------------------------------------------
// table pour la porte
SDoorElem   tabDoor[WALL_TABLE_HEIGHT][WALL_TABLE_WIDTH] =
{
    // Rang 3 (le plus loin)
    {
        {"", 0, 0, 0, 0, 0, 0},  // 03
        {"0",  29, 62,  30, 62,  72,  99},  // 13
        {"0",  90, 63,  90, 63, 133, 100},  // 23
        {"0", 151, 62, 151, 62, 193,  99},  // 33
        {"", 0, 0, 0, 0, 0, 0}   // 43
    },

    // Rang 2
    {
        {"", 0, 0, 0, 0, 0, 0},  // 02
        {"1",  -3, 56,   0, 56,  58, 114}, // 12
        {"1",  80, 57,  80, 57, 143, 115}, // 22
        {"1", 164, 56, 165, 56, 223, 114}, // 32
        {"", 0, 0, 0, 0, 0, 0}   // 42
    },

    // Rang 1
    {
        {"", 0, 0, 0, 0, 0, 0},  // 01
        {"2", -63, 51,   0, 51,  31, 136}, // 11
        {"2",  64, 49,  64, 49, 159, 134}, // 21
        {"2", 192, 51, 192, 51, 223, 136}, // 31
        {"", 0, 0, 0, 0, 0, 0}   // 41
    },

    // Rang 0 (le plus proche)
    {
        {"", 0, 0, 0, 0, 0, 0},  // 00
        {"", 0, 0, 0, 0, 0, 0},  // 10
        {"", 0, 0, 0, 0, 0, 0},  // 20
        {"", 0, 0, 0, 0, 0, 0},  // 30
        {"", 0, 0, 0, 0, 0, 0}   // 40
    }
};

//----------------------------------------------------------------------------
// table pour le bouton

struct SButtonElem
{
    int16_t x, y;
    float   scale;
};

SButtonElem   tabButton[WALL_TABLE_HEIGHT][WALL_TABLE_WIDTH] =
{
    // Rang 3 (le plus loin)
    {
        {0, 0, 0.0},  // 03
        {0, 0, 0.0},  // 13
        {137, 74, 0.43},  // 23
        {197, 72, 0.43},  // 33
        {0, 0, 0.0}   // 43
    },

    // Rang 2
    {
        {0, 0, 0.0},  // 02
        {0, 0, 0.0},  // 12
        {150, 75, 0.66}, // 22
        {0, 0, 0.0},  // 32
        {0, 0, 0.0}   // 42
    },

    // Rang 1
    {
        {0, 0, 0.0},  // 01
        {0, 0, 0.0},  // 11
        {167, 76,  1.0}, // 21
        {0, 0, 0.0},  // 31
        {0, 0, 0.0}   // 41
    },

    // Rang 0 (le plus proche)
    {
        {0, 0, 0.0},  // 00
        {0, 0, 0.0},  // 10
        {0, 0, 0.0},  // 20
        {0, 0, 0.0},  // 30
        {0, 0, 0.0}   // 40
    }
};

//----------------------------------------------------------------------------
void    CDoors::drawFrameElement(QImage* image, CVec2 tablePos, SDoorElem table[WALL_TABLE_HEIGHT][WALL_TABLE_WIDTH], bool isFlipped)
{
    SDoorElem&  cell = table[tablePos.y][tablePos.x];

    if (cell.file[0] != 0)
    {
        std::string fileName = std::string("gfx/3DView/doors/") + std::string(cell.file);
        QImage  elemImage = fileCache.getImage(fileName);
        CVec2 pos(cell.x, cell.y);
        QRect clip(cell.cx1, cell.cy1,
                   cell.cx2 - cell.cx1 + 1, cell.cy2 - cell.cy1 + 1);
        graph2D.drawImage(image, pos, elemImage, 0, isFlipped, clip);
    }
}

//----------------------------------------------------------------------------
void    CDoors::clipToMainView(QRect* clip)
{
    if (clip->y() < 33)
    {
        clip->setHeight(clip->height() - (33 - clip->y()));
        clip->setTop(33);
    }

    if (clip->x() + clip->width() > MAINVIEW_WIDTH)
        clip->setWidth(MAINVIEW_WIDTH - clip->x());
}

//----------------------------------------------------------------------------
void    CDoors::drawDoor(QImage* image, CVec2 tablePos, CTile* tile)
{
    SDoorElem&  cell = tabDoor[tablePos.y][tablePos.x];
    int doorNum = cell.file[0] - '0';

    if (doorNum >= 0)
    {
        //----------------------------------------------------------------------------
        // récupère l'image de la porte
        int doorType = tile->getDoorParam("Type");
        std::string fileName = std::string("gfx/3DView/doors/") + doorsDatas[doorType].files[doorNum].toUtf8().constData();
        QImage  doorImage = fileCache.getImage(fileName);

        //----------------------------------------------------------------------------
        // récupère l'image du décor
        int ornateType = tile->getDoorOrnateParam("Decor");
        std::string ornateName = doorOrnatesDatas[ornateType].file.toUtf8().constData();

        QImage  ornateImage;
        CVec2   ornatePos;
        if (ornateName.empty() == false)
        {
            // réduit le décor, l'assombrit, et le dessine sur l'image de la porte
            ornateImage = fileCache.getImage(std::string("gfx/3DView/door_ornates/") + ornateName);
            static const float shadowLevels[] = {0.0f, 0.2f, 0.4f};
            float shadow = shadowLevels[WALL_TABLE_HEIGHT - 2 - tablePos.y];
            QImage originalOrnate = ornateImage;
            graph2D.darken(&ornateImage, shadow);

            ornatePos = doorOrnatesDatas[ornateType].pos;
            float scale = doorImage.width() / 96.0;
            ornatePos *= scale;

            graph2D.drawImageScaled(&doorImage, ornatePos, ornateImage, scale);

            // décors avec transparence
            if (ornateType == DOOR_ORNATE_ARCHED || ornateType == DOOR_ORNATE_ARCHED + 1)
            {
                QSize   qSize = originalOrnate.size();

                for (int y = 0; y <= qSize.height(); ++y)
                    for (int x = 0; x <= qSize.width(); ++x)
                    {
                        QRgb pixel = originalOrnate.pixel(x, y);

                        if (qRed(pixel) == 204 &&
                            qGreen(pixel) == 136 &&
                            qBlue(pixel) == 102)
                        {
                            CVec2   destPos = ornatePos + CVec2(x, y) * scale;
                            doorImage.setPixel(destPos.x, destPos.y, 0);
                        }
                    }
            }
        }

        //----------------------------------------------------------------------------
        // dessin
        bool isHorizontal = tile->getBoolParam("estHorizontale");
        int doorPos = tile->getIntParam("pos") / DOOR_POS_FACTOR;

        if (isHorizontal == false)
        {
            //----------------------------------------------------------------------------
            // porte verticale
            static const int doorScalesY[] = {88, 61, 38};
            float scale = doorScalesY[WALL_TABLE_HEIGHT - 2 - tablePos.y];
            doorPos = (doorPos * scale) / 88;

            CVec2 pos(cell.x, cell.y + doorPos);
            QRect clip(cell.cx1, cell.cy1,
                       cell.cx2 - cell.cx1 + 1, cell.cy2 - cell.cy1 + 1);
            clipToMainView(&clip);

            if (clip.width() > 0 && clip.height() > 0)
                graph2D.drawImage(image, pos, doorImage, 0, false, clip);
        }
        else
        {
            //----------------------------------------------------------------------------
            // porte horizontale
            static const int doorScalesX[] = {96, 64, 44};
            float scale = doorScalesX[WALL_TABLE_HEIGHT - 2 - tablePos.y];
            doorPos = (doorPos * scale) / 96;

            CVec2 pos(cell.x + doorPos, cell.y);
            int halfwidth = (cell.cx2 - cell.cx1 + 1) / 2;

            QRect clip1(cell.cx1, cell.cy1,
                       halfwidth + doorPos, cell.cy2 - cell.cy1 + 1);
            clipToMainView(&clip1);

            if (clip1.width() > 0 && clip1.height() > 0)
                graph2D.drawImage(image, pos, doorImage, 0, false, clip1);

            pos.x = cell.x - doorPos;
            QRect clip2(cell.cx1 + halfwidth - doorPos, cell.cy1,
                       halfwidth + doorPos, cell.cy2 - cell.cy1 + 1);
            clipToMainView(&clip2);

            if (clip2.width() > 0 && clip2.height() > 0)
                graph2D.drawImage(image, pos, doorImage, 0, false, clip2);
        }
    }
}

//----------------------------------------------------------------------------
bool CDoors::isFacing(CTile* tile)
{
    EWallSide playerSide = player.getWallSide(eWallSideUp);

    return (tile->mWalls[playerSide].getType() != eWallDoorSide);
}

//----------------------------------------------------------------------------
void CDoors::draw(QImage* image, CVec2 mapPos, CVec2 tablePos)
{
    CTile*  tile = map.getTile(mapPos);

    if (isFacing(tile) == true)
    {
        //-------------------------------------------------------------------------------------
        // dessine l'encadrement
        drawFrameElement(image, tablePos, tabTopFrame, false);
        drawFrameElement(image, tablePos, tabLeftFrame, false);
        drawFrameElement(image, tablePos, tabRightFrame, true);

        //-------------------------------------------------------------------------------------
        // dessine le bouton
        if (tile->getBoolParam("aUnBouton") == true)
        {
            SButtonElem& cell = tabButton[tablePos.y][tablePos.x];
            CVec2 pos(cell.x, cell.y);
            float scale = cell.scale;

            if (scale != 0.0f)
            {
                QImage  button = fileCache.getImage("gfx/3DView/doors/Button.png");
                static const float shadowLevels[] = {0.0f, 0.2f, 0.4f};
                float shadow = shadowLevels[WALL_TABLE_HEIGHT - 2 - tablePos.y];

                graph2D.darken(&button, shadow);
                graph2D.drawImageScaled(image, pos, button, scale);

                if (tablePos == CVec2(2, 2))
                {
                    CRect   rect(pos,
                                 CVec2(pos.x + button.width() - 1, pos.y + button.height() - 1));
                    mouse.addArea(eMouseAreaG_DoorButton, rect, eCursor_Hand, (void*)mapPos.x, (void*)mapPos.y);
                }
            }
        }

        //-------------------------------------------------------------------------------------
        // dessine la porte avec le décor
        drawDoor(image, tablePos, tile);
    }
}

//----------------------------------------------------------------------------
// Met à jour les animations des portes
void CDoors::update()
{
    for (int y = 0; y < map.mSize.y; ++y)
        for (int x = 0; x < map.mSize.x; ++x)
        {
            CVec2   mapPos = CVec2(x, y);
            CTile*  tile = map.getTile(mapPos);

            if (tile->getType() == eTileDoor)
            {
                bool isHorizontal = tile->getBoolParam("estHorizontale");
                int animLenght = (isHorizontal == false ? 88 : 48) * DOOR_POS_FACTOR;
                int pos = tile->getIntParam("pos");
                bool isOpened = tile->getBoolParam("estOuverte");

                if (isOpened == false)
                {
                    if (pos < 0)
                    {
                        tile->setBoolParam("estEnMouvement", true);
                        sound->play(mapPos, "sound/Door.wav", 0);
                        int time = DOOR_ANIMATION_TIME * CTimer::getGameTimeFactor();

                        if (time != 0)
                        {
                            pos += animLenght / time;
                            tile->setIntParam("pos", pos);
                        }
                    }

                    if (pos >= 0)
                    {
                        tile->setBoolParam("estEnMouvement", false);
                        sound->stop(mapPos, "sound/Door.wav");
                        tile->setIntParam("pos", 0);
                    }
                }
                else
                {
                    if (pos > -animLenght)
                    {
                        tile->setBoolParam("estEnMouvement", true);
                        sound->play(mapPos, "sound/Door.wav", 0);
                        int time = DOOR_ANIMATION_TIME * CTimer::getGameTimeFactor();

                        if (time != 0)
                        {
                            pos -= animLenght / time;
                            tile->setIntParam("pos", pos);
                        }
                    }

                    if (pos <= -animLenght)
                    {
                        tile->setBoolParam("estEnMouvement", false);
                        sound->stop(mapPos, "sound/Door.wav");
                        tile->setIntParam("pos", -animLenght);
                    }
                }
            }
        }
}

//---------------------------------------------------------------------------------------------
void    CDoors::readDoorsDB()
{
    QDomDocument doc;

    QFile f("databases/doors.xml");
    f.open(QIODevice::ReadOnly);
    doc.setContent(&f);
    f.close();

    QDomElement root = doc.documentElement();
    QDomElement door = root.firstChild().toElement();

    while(!door.isNull())
    {
        CDoorData   newDoor;

        if (door.tagName() == "door")
        {
            QDomElement doorInfo = door.firstChild().toElement();

            while(!doorInfo.isNull())
            {
                if (doorInfo.tagName() == "image1")
                {
                    newDoor.files[0] = doorInfo.text();
                }
                else if (doorInfo.tagName() == "image2")
                {
                    newDoor.files[1] = doorInfo.text();
                }
                else if (doorInfo.tagName() == "image3")
                {
                    newDoor.files[2] = doorInfo.text();
                }

                doorInfo = doorInfo.nextSibling().toElement();
            }
            doorsDatas.push_back(newDoor);
        }
        door = door.nextSibling().toElement();
    }
}

//---------------------------------------------------------------------------------------------
void    CDoors::readDoorOrnatesDB()
{
    QDomDocument doc;

    QFile f("databases/door_ornates.xml");
    f.open(QIODevice::ReadOnly);
    doc.setContent(&f);
    f.close();

    QDomElement root = doc.documentElement();
    QDomElement ornate = root.firstChild().toElement();

    while(!ornate.isNull())
    {
        CDoorOrnateData   newOrnate;

        if (ornate.tagName() == "door_ornate")
        {
            QDomElement ornateInfo = ornate.firstChild().toElement();

            while(!ornateInfo.isNull())
            {
                if (ornateInfo.tagName() == "image")
                {
                    newOrnate.file = ornateInfo.text();
                }
                else if (ornateInfo.tagName() == "pos")
                {
                    newOrnate.pos.x = ornateInfo.attribute("x").toInt();
                    newOrnate.pos.y = ornateInfo.attribute("y").toInt();
                }

                ornateInfo = ornateInfo.nextSibling().toElement();
            }
            doorOrnatesDatas.push_back(newOrnate);
        }
        ornate = ornate.nextSibling().toElement();
    }
}

//----------------------------------------------------------------------------
bool CDoors::isOpened(CTile* tile)
{
    bool isOpened = tile->getBoolParam("estOuverte");
    bool isMoving = tile->getBoolParam("estEnMouvement");

    if (isOpened == true && isMoving == false)
        return true;
    else
        return false;
}
