﻿#include <QApplication>
#include <QQmlApplicationEngine>
#include <QQmlContext>
#include <QDebug>
#include <time.h>
#include "../common/sources/Map.h"
#include "bridge.h"
#include "keyboard.h"
#include "player.h"
#include "tiles.h"
#include "walls.h"
#include "doors.h"
#include "interface.h"
#include "character.h"
#include "objects.h"
#include "game.h"
#include "sound.h"

QQmlContext     *ctxt;
CSoundManager   *sound;

int main(int argc, char *argv[])
{
    srand(time(NULL));
    QApplication app(argc, argv);

    QQmlApplicationEngine engine;

    // lit les bases de données
    interface.readTextDB();
    tiles.readFloorOrnatesDB();
    tiles.readTilesDB();
    walls.readOrnatesDB();
    walls.readWallsDB();
    doors.readDoorsDB();
    doors.readDoorOrnatesDB();
    CCharacter::readCharactersDB();
    objects.readObjectsDB();

    // enregistre nos fonctions
    ctxt = engine.rootContext();
    ctxt->setContextProperty("bridge", &bridge);

    // enregistre l'ImageProvider
    engine.addImageProvider("myimageprovider", new MyImageProvider);

    // charge la map
    game.loadLevel(1, CVec2(1, 3), 2);

    sound = new CSoundManager;

    // lance l'interface
    engine.load(QUrl(QStringLiteral("qrc:/main.qml")));

	// enregistre l'EventFilter pour le clavier
    QApplication::instance()->installEventFilter(&keyboard);

    int retValue = app.exec();

	// désenregistre l'EventFilter
    QApplication::instance()->removeEventFilter(&keyboard);

    return retValue;
}

