﻿#include "tiles.h"
#include <QtXml/qdom.h>
#include <QFile>
#include "../common/sources/system/filecache.h"
#include "../common/sources/system/cgraphics2d.h"
#include "game.h"
#include <QDebug>
#include "player.h"
#include "interface.h"
#include "scripts.h"

CTiles   tiles;

//---------------------------------------------------------------------------------------------
CTiles::CTiles()
{
    pressPlateStates = NULL;
}

//---------------------------------------------------------------------------------------------
void CTiles::initMap()
{
    allocPressPlates();
    for (int i = 0; i < map.mSize.x * map.mSize.y; ++i)
        pressPlateStates[i] = false;
}

//---------------------------------------------------------------------------------------------
CTiles::~CTiles()
{
    freePressPlates();
}

//-----------------------------------------------------------------------------------------
void CTiles::freePressPlates()
{
    if (pressPlateStates != NULL)
        delete[] pressPlateStates;
}

//-----------------------------------------------------------------------------------------
void CTiles::allocPressPlates()
{
    freePressPlates();
    int mapSize = map.mSize.x * map.mSize.y;
    pressPlateStates = new bool[mapSize];
}

//---------------------------------------------------------------------------------------------
void    CTiles::drawFloorOrnate(QImage* image, CVec2 tablePos, int ornate)
{
    COrnateData ornateData = ornatesDatas[ornate];
    bool    flip = false;

    if (tablePos.x >= (int)((WALL_TABLE_WIDTH + 1) / 2))
    {
        tablePos.x = WALL_TABLE_WIDTH - 1 - tablePos.x;
        flip = true;
    }

    COrnateElement  elem = ornateData.table[tablePos.y][tablePos.x];

    if (elem.image.isEmpty() == false)
    {
        QImage  ornateImage = fileCache.getImage(elem.image.toLocal8Bit().constData());
        CVec2   pos = elem.pos;

        if (flip == true)
            pos.x = MAINVIEW_WIDTH - pos.x - ornateImage.width();

        graph2D.drawImage(image, pos, ornateImage, 0, flip);
    }
}

//--------------------------------------------------------------------------------------------------------
void    CTiles::readFloorOrnatesDB()
{
    QDomDocument doc;

    QFile f("databases/floor_ornates.xml");
    f.open(QIODevice::ReadOnly);
    doc.setContent(&f);
    f.close();

    QDomElement root = doc.documentElement();
    QDomElement ornate = root.firstChild().toElement();

    while(!ornate.isNull())
    {
        if (ornate.tagName() == "floor_ornate")
        {
            COrnateData   newOrnate;

            newOrnate.name = ornate.attribute("name");

            QDomElement ornateInfo = ornate.firstChild().toElement();

            while(!ornateInfo.isNull())
            {
                QString tag = ornateInfo.tagName();

                if (tag.startsWith("image_") == true)
                {
                    std::string str = tag.toUtf8().constData();
                    int x = str.at(str.size() - 2) - '0';
                    int y = str.at(str.size() - 1) - '0';
                    COrnateElement* elem = &newOrnate.table[y][x];
                    elem->image = QString("gfx/3DView/floor_ornates/") + ornateInfo.text();
                }
                else if (tag.startsWith("pos_") == true)
                {
                    std::string str = tag.toUtf8().constData();
                    int x = str.at(str.size() - 2) - '0';
                    int y = str.at(str.size() - 1) - '0';
                    COrnateElement* elem = &newOrnate.table[y][x];
                    elem->pos.x = ornateInfo.attribute("x").toInt();
                    elem->pos.y = ornateInfo.attribute("y").toInt();
                }

                ornateInfo = ornateInfo.nextSibling().toElement();
            }
            ornatesDatas.push_back(newOrnate);
        }
        ornate = ornate.nextSibling().toElement();
    }
}

//---------------------------------------------------------------------------------------------
void    CTiles::readTilesDB()
{
    QDomDocument doc;

    QFile f("databases/tiles.xml");
    f.open(QIODevice::ReadOnly);
    doc.setContent(&f);
    f.close();

    QDomElement root = doc.documentElement();
    QDomElement tile = root.firstChild().toElement();

    while(!tile.isNull())
    {
        std::vector<CParamType> paramsTypesList;

        if (tile.tagName() == "tile")
        {
            QDomElement tileInfo = tile.firstChild().toElement();

            while(!tileInfo.isNull())
            {
                if (tileInfo.tagName() == "param")
                {
                    QString paramType = tileInfo.attribute("type");
                    CParamType  newParamType;

                    if (paramType == "ornate")
                    {
                        newParamType.mType = eParamOrnate;
                        newParamType.mName = tileInfo.text();
                        paramsTypesList.push_back(newParamType);
                    }
                    else if (paramType == "int")
                    {
                        newParamType.mType = eParamInt;
                        newParamType.mName = tileInfo.text();
                        paramsTypesList.push_back(newParamType);
                    }
                    else if (paramType == "bool")
                    {
                        newParamType.mType = eParamBool;
                        newParamType.mName = tileInfo.text();
                        paramsTypesList.push_back(newParamType);
                    }
                    else if (paramType == "door")
                    {
                        newParamType.mType = eParamDoor;
                        newParamType.mName = tileInfo.text();
                        paramsTypesList.push_back(newParamType);
                    }
                    else if (paramType == "door_ornate")
                    {
                        newParamType.mType = eParamDoorOrnate;
                        newParamType.mName = tileInfo.text();
                        paramsTypesList.push_back(newParamType);
                    }
                    else if (paramType == "enum")
                    {
                        newParamType.mType = eParamEnum;
                        newParamType.mName = tileInfo.text();
                        paramsTypesList.push_back(newParamType);
                    }
                    else if (paramType == "script")
                    {
                        newParamType.mType = eParamScript;
                        newParamType.mName = tileInfo.text();
                        paramsTypesList.push_back(newParamType);
                    }
                }

                tileInfo = tileInfo.nextSibling().toElement();
            }
            CMap::mTilesParams.push_back(paramsTypesList);
        }
        tile = tile.nextSibling().toElement();
    }
}

//----------------------------------------------------------------------------
// Met à jour les plaques de pression et exécute leurs scripts
void CTiles::updatePressPlates()
{
    for (int y = 0; y < map.mSize.y; ++y)
        for (int x = 0; x < map.mSize.x; ++x)
        {
            CVec2   pos(x, y);
            CTile*  tile = map.getTile(pos);

            bool&   state = pressPlateStates[y * map.mSize.x + x];
            bool    lastState = state;
            state = false;

            if (tile->getType() == eTilePressPlate)
            {
                if (tile->getBoolParam("parChampions") == true && interface.isChampionEmpty(0) == false && player.pos == pos)
                {
                    state = true;
                }
                else if (tile->getBoolParam("parObjets") == true)
                {
                    CVec2   pos2 = pos * 2;

                    if (map.findObjectsStack(pos2, eWallSideMax) != NULL ||
                        map.findObjectsStack(pos2 + CVec2(1, 0), eWallSideMax) != NULL ||
                        map.findObjectsStack(pos2 + CVec2(0, 1), eWallSideMax) != NULL ||
                        map.findObjectsStack(pos2 + CVec2(1, 1), eWallSideMax) != NULL)
                    {
                        state = true;
                    }
                }

                // marche sur une plaque de pression
                if (lastState == false && state == true)
                {
                    if (tile->getBoolParam("playSound") == true)
                        sound->play(pos, "sound/Switch.wav");

                    scripts.execute(tile->getScriptParam("siPressee"));
                }

                // quitte une plaque de pression
                if (lastState == true && state == false)
                    scripts.execute(tile->getScriptParam("siRelachee"));
            }
        }
}

//----------------------------------------------------------------------------
struct SStairsData
{
    char        file[32];
    uint16_t    x, y;
};

// table pour les escaliers descendant, en face
static const SStairsData   gDownFacingStairs[WALL_TABLE_HEIGHT][WALL_TABLE_WIDTH] =
{
    {{"", 0, 0}, {"Down_13_F.png", 8, 63}, {"Down_23_F.png", 75, 58}, {"Down_13_F.png", 141, 63}, {"", 0, 0}}, // Row 3
    {{"", 0, 0}, {"Down_12_F.png", 0, 57}, {"Down_22_F.png", 63, 57}, {"Down_12_F.png", 163, 57}, {"", 0, 0}}, // Row 2
    {{"", 0, 0}, {"Down_11_F.png", 0, 51}, {"Down_21_F.png", 35, 50}, {"Down_11_F.png", 192, 51}, {"", 0, 0}}, // Row 1
    {{"", 0, 0}, {"", 0, 0},               {"", 0, 0},                {"", 0, 0},                 {"", 0, 0}}  // Row 0
};

// table pour les escaliers montant, en face
static const SStairsData   gUpFacingStairs[WALL_TABLE_HEIGHT][WALL_TABLE_WIDTH] =
{
    {{"", 0, 0}, {"Up_13_F.png", 14, 59}, {"Up_23_F.png", 78, 58}, {"Up_13_F.png", 147, 59}, {"", 0, 0}}, // Row 3
    {{"", 0, 0}, {"Up_12_F.png",  0, 53}, {"Up_22_F.png", 62, 53}, {"Up_12_F.png", 164, 53}, {"", 0, 0}}, // Row 2
    {{"", 0, 0}, {"Up_11_F.png",  0, 42}, {"Up_21_F.png", 32, 42}, {"Up_11_F.png", 192, 42}, {"", 0, 0}}, // Row 1
    {{"", 0, 0}, {"", 0, 0},              {"", 0, 0},              {"", 0, 0},               {"", 0, 0}}  // Row 0
};

// table pour les escaliers descendant, de coté
static const SStairsData   gDownSideStairs[WALL_TABLE_HEIGHT][WALL_TABLE_WIDTH] =
{
    {{"", 0, 0}, {"", 0, 0},                      {"", 0, 0}, {"", 0, 0},                       {"", 0, 0}}, // Row 3
    {{"", 0, 0}, {"Side_12_F.png",      60,  89}, {"", 0, 0}, {"Side_12_F.png",      156,  89}, {"", 0, 0}}, // Row 2
    {{"", 0, 0}, {"Down_Side_11_F.png", 32,  95}, {"", 0, 0}, {"Down_Side_11_F.png", 172,  95}, {"", 0, 0}}, // Row 1
    {{"", 0, 0}, {"Side_10_F.png",       0, 106}, {"", 0, 0}, {"Side_10_F.png",      208, 106}, {"", 0, 0}}  // Row 0
};

// table pour les escaliers montant, de coté
static const SStairsData   gUpSideStairs[WALL_TABLE_HEIGHT][WALL_TABLE_WIDTH] =
{
    {{"", 0, 0}, {"", 0, 0},                    {"", 0, 0}, {"", 0, 0},                     {"", 0, 0}}, // Row 3
    {{"", 0, 0}, {"Side_12_F.png",    60,  89}, {"", 0, 0}, {"Side_12_F.png",    156,  89}, {"", 0, 0}}, // Row 2
    {{"", 0, 0}, {"Up_Side_11_F.png", 32,  91}, {"", 0, 0}, {"Up_Side_11_F.png", 172,  91}, {"", 0, 0}}, // Row 1
    {{"", 0, 0}, {"Side_10_F.png",     0, 106}, {"", 0, 0}, {"Side_10_F.png",    208, 106}, {"", 0, 0}}  // Row 0
};

void CTiles::drawStairs(QImage* image, CTile* tile, CVec2 tablePos)
{
    const SStairsData*  data;

    // trouve le coté des escaliers
    EWallSide   side;
    int i;

    for (i = 0; i < 4; ++i)
        if (tile->mWalls[i].getType() == eWallStairs)
            break;
    side = (EWallSide)i;
    bool isDown = (tile->getIntParam("Niveau") > game.currentLevel);

    if (side != player.getWallSide(eWallSideDown))
    {
        // récupère les données de la table
        if (side == player.getWallSide(eWallSideUp))
        {
            // escaliers en face
            if (isDown)
                data = &gDownFacingStairs[tablePos.y][tablePos.x];
            else
                data = &gUpFacingStairs[tablePos.y][tablePos.x];
        }
        else
        {
            // escaliers de coté
            if (isDown)
                data = &gDownSideStairs[tablePos.y][tablePos.x];
            else
                data = &gUpSideStairs[tablePos.y][tablePos.x];
        }

        // affichage
        if (data->file[0] != 0)
        {
            std::string fileName = std::string("gfx/3DView/stairs/") + std::string(data->file);
            CVec2   pos = CVec2(data->x, data->y);
            bool    flip = (tablePos.x > 2);
            QImage  stairsImage = fileCache.getImage(fileName);
            graph2D.drawImage(image, pos, stairsImage, 0, flip);
        }
    }
    else if (tablePos == CVec2(2, 3))
    {
        // cas spécial si on est sur la case de l'escalier
        if (isDown)
        {
            QImage  stairsImage = fileCache.getImage("gfx/3DView/stairs/Down_20_L.png");
            graph2D.drawImage(image, CVec2(0, 109), stairsImage);
            graph2D.drawImage(image, CVec2(194, 109), stairsImage, 0, true);
        }
        else
        {
            QImage  stairsImage = fileCache.getImage("gfx/3DView/stairs/Up_20_L.png");
            graph2D.drawImage(image, CVec2(0, 91), stairsImage);
            graph2D.drawImage(image, CVec2(194, 91), stairsImage, 0, true);
        }
    }
}

//----------------------------------------------------------------------------
bool* CTiles::getPressPlates()
{
    return pressPlateStates;
}

//----------------------------------------------------------------------------
void CTiles::setPressPlates(bool* table)
{
    allocPressPlates();
    for (int i = 0; i < map.mSize.x * map.mSize.y; ++i)
        pressPlateStates[i] = table[i];
}
